//-----------------------------------------------------------------------------
// File: Lighting - Shadow.fx
// The first 8 lights (PIXELLIGHTS) are Pixel lights.  
// Spotlights support shadows.
// Copyright Tall Studios Ltd, 2014.  All rights reserved.
//---------------------------------------------------------------

#define MAXLIGHTS 8
#define PIXELLIGHTS 8
#define SMAP_SIZE 512
#define SHADOW_EPSILON 0.0005f

// Ambient Lighting Settings
float4		colourAmbient : Ambient0;
float		fAmbientIntensity;

//Lights
const float3 HemisphericColor = float3( 0.15f, 0.2f, 0.30f);
const float3 Up = float3(0,1,0); //for the skylight
	
struct LIGHT
{
    float	iType;						// Type of light.  0=Directional, 1=Point, 2=Spotlight
    float4	vPosition: POSITION;		// Position of light (Point or Spotlight only)
    float4	vDirection;					// Direction of light (Directional or Spotlight only)
    float4	colour: COLOR;				// Colour of light
    float	fRange;						// Range of light
    float	fInnerAngle;				// Inner Angle of spotlight (Spotlight only)
    float	fOuterAngle;				// Outer Angle of spotlight (Spotlight only)
    bool	bEnabled;					// Is this light switched on?
	bool	bShadowActive;				// Does the light cast a Shadow
};

uniform LIGHT	light[MAXLIGHTS];
uniform int		iNumberOfLights = 0;

uniform float SpecPow <
	string UIName = "Specular Glossiness";
	float UIMin = 0.0f;
	float UIMax = 100.0f;
	float UIStep = 0.1f;
>   = 30.0f;

texture  textureShadow;

sampler2D g_samShadow =
sampler_state
{
    Texture = <textureShadow>;
    MinFilter = Point;
    MagFilter = Point;
    MipFilter = Point;
    AddressU = Clamp;
    AddressV = Clamp;
};


//Function: Omni Lighting + Attenuation + Specular
static inline float4 OmniAttenuationSpecular(in int lightnumber, in float3 WorldPosition, float3 Normal, float3 ViewDir)
{ 
	float3 LightDir = light[lightnumber].vPosition-WorldPosition;
	half dist = length(LightDir);	
	LightDir = normalize(LightDir);
	half attenuation = 1-saturate(dist/light[lightnumber].fRange);					//Attenuation
	attenuation = attenuation * attenuation;										//More realistic attenuation
	float3 halfVector = normalize(LightDir + ViewDir);
	float Specular = pow(saturate(dot(Normal,halfVector)),SpecPow)*attenuation;		//Specular term
	float3 Lighting = saturate(dot(Normal,LightDir))*attenuation*light[lightnumber].colour;
	return float4(Lighting,Specular);
}
   
//Function: Directional Lighting + Specular
static inline float4 DirectionalSpecular(int lightnumber, float3 Normal, float3 ViewDir)
{
	float3 LightDir = normalize(light[lightnumber].vDirection);
	float3 halfVector = normalize(LightDir + ViewDir);
	float Specular = pow(saturate(dot(Normal,halfVector)),SpecPow);					//Specular term
	float3 Lighting = saturate(dot(Normal,LightDir))*light[lightnumber].colour;
	return float4(Lighting,Specular);
}
   
//Function: Omni Lighting + Attenuation + Specular
static inline float4 SpotAttenuationSpecularShadow(in int lightnumber, in float3 WorldPosition, float3 Normal, float3 ViewDir, float4 PosLightSpace)
{ 
	float3 LightDir = light[lightnumber].vPosition-WorldPosition;
	half dist = length(LightDir);	
	LightDir = normalize(LightDir);

	if(dot(LightDir, light[lightnumber].vDirection) > cos(light[lightnumber].fOuterAngle))
	{
        //transform from RT space to texture space.
        float2 ShadowTexC = 0.5 * PosLightSpace.xy / PosLightSpace.w + float2( 0.5f, 0.5f );
        ShadowTexC.y = 1.0f - ShadowTexC.y;

        // transform to texel space
        float2 texelpos = SMAP_SIZE * ShadowTexC;
        
        // Determine the lerp amounts
        float2 lerps = frac( texelpos );

        //read in bilerp stamp, doing the shadow checks
        float sourcevals[4];
        sourcevals[0] = (tex2D( g_samShadow, ShadowTexC ) + SHADOW_EPSILON < PosLightSpace.z / PosLightSpace.w)? 0.0f: 1.0f;
        sourcevals[1] = (tex2D( g_samShadow, ShadowTexC + float2(1.0/SMAP_SIZE, 0) ) + SHADOW_EPSILON < PosLightSpace.z / PosLightSpace.w)? 0.0f: 1.0f;
        sourcevals[2] = (tex2D( g_samShadow, ShadowTexC + float2(0, 1.0/SMAP_SIZE) ) + SHADOW_EPSILON < PosLightSpace.z / PosLightSpace.w)? 0.0f: 1.0f;
        sourcevals[3] = (tex2D( g_samShadow, ShadowTexC + float2(1.0/SMAP_SIZE, 1.0/SMAP_SIZE) ) + SHADOW_EPSILON < PosLightSpace.z / PosLightSpace.w)? 0.0f: 1.0f;

        // lerp between the shadow values to calculate our light amount
        float LightAmount = lerp( lerp( sourcevals[0], sourcevals[1], lerps.x ),
                                  lerp( sourcevals[2], sourcevals[3], lerps.x ),
                                  lerps.y );
        // Light it
		//float4 Diffuse;
		//Diffuse = ( saturate( dot( LightDir, normalize( Normal ) ) ) * LightAmount );
		//return Diffuse;

		half attenuation = LightAmount-saturate(dist/light[lightnumber].fRange);					//Attenuation
		attenuation = attenuation * attenuation;										//More realistic attenuation
	   
		half coneDot = dot(normalize(light[lightnumber].vDirection),LightDir);			//Spot cone attenuation - Will be optimised without cos
		attenuation *= saturate((coneDot-cos(light[lightnumber].fOuterAngle))/(cos(light[lightnumber].fInnerAngle)-cos(light[lightnumber].fOuterAngle)));
		
		float3 halfVector = normalize(LightDir + ViewDir);
		float Specular = pow(saturate(dot(Normal,halfVector)),SpecPow)*attenuation;		//Specular term
		float3 Lighting = saturate(dot(Normal,LightDir))*attenuation*light[lightnumber].colour;
		return float4(Lighting,Specular);

	}
	return float4(0.f,0.f,0.f,0.f);
}

//Function: Omni Lighting + Attenuation + Specular
static inline float4 SpotAttenuationSpecular(in int lightnumber, in float3 WorldPosition, float3 Normal, float3 ViewDir)
{ 
	float3 LightDir = light[lightnumber].vPosition-WorldPosition;
	half dist = length(LightDir);	
	LightDir = normalize(LightDir);
	half attenuation = 1-saturate(dist/light[lightnumber].fRange);					//Attenuation
	attenuation = attenuation * attenuation;										//More realistic attenuation
	   
	half coneDot = dot(normalize(light[lightnumber].vDirection),LightDir);			//Spot cone attenuation - Will be optimised without cos
	attenuation *= saturate((coneDot-cos(light[lightnumber].fOuterAngle))/(cos(light[lightnumber].fInnerAngle)-cos(light[lightnumber].fOuterAngle)));
		
	float3 halfVector = normalize(LightDir + ViewDir);
	float Specular = pow(saturate(dot(Normal,halfVector)),SpecPow)*attenuation;		//Specular term
	float3 Lighting = saturate(dot(Normal,LightDir))*attenuation*light[lightnumber].colour;
	return float4(Lighting,Specular);
}

   
float4x4	matWorldViewProj : WorldViewProjection;
float4x4 matWorld : World;
float4x4 matWorldView : WorldView;
float4x4 matViewToLightProj;
float4 vCameraPosition : WORLDCAMERAPOSITION;


uniform bool b_UseNormal <
	string UIName = "Use Normal map";
> = false;
uniform bool b_UseReflection <
	string UIName = "Use Reflection map";
> = false;
uniform bool b_UseAmbient <
	string UIName = "Use ambient colour";
> = true;
uniform bool b_UseDiffuse <
	string UIName = "Use Diffuse map";
> = true;
uniform bool b_UseSpecMask <
	string UIName = "Use Specular map";
> = false;
uniform bool b_UseLuminosity <
	string UIName = "Use Luminosity map";
> = false;


texture diffusemap : DIFFUSEMAP <
	string name = "";
	string UIName = "Diffuse Texture";
	int Texcoord = 0;
	int MapChannel = 1;
	>;
	
sampler2D Diffusemap = sampler_state
{
	   Texture   = (diffusemap);
	   MinFilter = LINEAR;
	   MagFilter = LINEAR;
	   MipFilter = LINEAR;   
};

float4 DiffuseColor : COLOR <
	string UIName = "DiffuseColor";
	string UIWidget = "Color";
> = float4( 0.7,0.7,0.7, 1.0);  

	
texture normalmap : NORMALMAP <
	string name = "";
	string UIName = "Normal Texture";
	int Texcoord = 0;
	int MapChannel = 1;
	>;

// Sampler with bilinear filtering
sampler2D Normalmap = sampler_state
{
	   Texture   = (normalmap);
	   MinFilter = LINEAR;
	   MagFilter = LINEAR;
	   MipFilter = LINEAR;   
};

texture specularmap : SPECULARMAP <
	string name = "";
	string UIName = "Specular Texture";
	int Texcoord = 0;
	int MapChannel = 1;
	>;

// Sampler with bilinear filtering
sampler2D Specularmap = sampler_state
{
	   Texture   = (specularmap);
	   MinFilter = LINEAR;
	   MagFilter = LINEAR;
	   MipFilter = LINEAR;   
};

texture envcubemap : ENVCUBEMAP <
	string name = "";
	string UIName = "Cubemap Texture";
	string Type = "Cube";
	>;

// Sampler with bilinear filtering
samplerCUBE EnvCubemap = sampler_state
{
	   Texture   = (envcubemap);
	   MinFilter = LINEAR;
	   MagFilter = LINEAR;
	   MipFilter = LINEAR;   
};

texture luminositymap : LUMINOSITYMAP <
	string name = "";
	string UIName = "Luminosity Texture";
	int Texcoord = 0;
	int MapChannel = 1;
	>;

// Sampler with bilinear filtering
sampler2D Luminositymap = sampler_state
{
	   Texture   = (luminositymap);
	   MinFilter = LINEAR;
	   MagFilter = LINEAR;
	   MipFilter = LINEAR;   
};


float3 SpecMask : COLOR <
	string UIName = "Specular Intensity";
	string UIWidget = "Color";
> = float3( 1.0,1.0,1.0);  

uniform float fresnelPower <
	string UIName = "Fresnel power";
	float UIMin = 0.0f;
	float UIMax = 100.0f;
	float UIStep = 0.1f;
>   = 1.0f;

uniform bool bFlipX <
	string UIName = "Flip Texture X";
> = false;
uniform bool bFlipY <
	string UIName = "Flip Texture Y";
> = false;

// -------------------------------------------------------------
// vertex shader output channels
// -------------------------------------------------------------
struct VS_OUTPUT
{
   float4 Pos: POSITION;
   float2 UV0 : TEXCOORD0;
   float3 ViewDir : TEXCOORD1;
   float3 Normal : TEXCOORD2;
   float3 Tangent : TEXCOORD3;
   float3 BiNormal : TEXCOORD4;
   float3 WorldPos : TEXCOORD5;
   float4 OtherLights : TEXCOORD6;
   float4 PosLightSpace : TEXCOORD7;
};



//-------------------------------------------------------------
// Vertex Shader Function (input channels)
//-------------------------------------------------------------

VS_OUTPUT VS( float4 Pos : POSITION, float3 Normal : NORMAL, float3 Tangent : TANGENT, float3 Binormal : BINORMAL, float2 UV0:TEXCOORD0 )
{
	VS_OUTPUT Out = (VS_OUTPUT) 0; 
   
	float4 WorldViewPos =   mul(Pos, matWorldViewProj); // transform Position 
	float4 WorldPos = mul(Pos, matWorld);
	float4 ShadowPos = mul(Pos, matWorldView);
	float3 ViewDir = vCameraPosition - WorldPos;
	float4 PosLightSpace = mul(ShadowPos, matViewToLightProj);
	Out.ViewDir = ViewDir;
          
	Out.Pos = WorldViewPos; // transform Position
	Out.WorldPos = WorldPos;
	Out.PosLightSpace = PosLightSpace;
   
	Out.Normal = mul(Normal,matWorld);
	Out.Tangent = mul(Tangent, matWorld);
	Out.BiNormal = mul(Binormal, matWorld);
   
	Out.UV0 = UV0;
	if(bFlipX)
		Out.UV0.x = -UV0.x;
	if(bFlipY)
		Out.UV0.y = -UV0.y;
	float3 WorldNormal = normalize(mul(Normal, matWorld)); 
         

	//vertex lights
	float4 TotalDiffuse	= 0;
	float3 Diffuse = 0;
	float4 VXLightDirection = 0;

	//Directional lights
	half d = 0;
	float3 LightDir;
	half dist;
	half attenuation;
	half coneDot;
	ViewDir = normalize(ViewDir);

	for (int i = PIXELLIGHTS; i < iNumberOfLights; i++)
	{    
		if (light[i].iType==0)
		{ 	
			TotalDiffuse += DirectionalSpecular(i, WorldNormal, ViewDir);	
		}
		if (light[i].iType==1)
		{ 	
			TotalDiffuse += OmniAttenuationSpecular(i, WorldPos, WorldNormal, ViewDir);
		}
		if (light[i].iType==2)
		{ 	
			TotalDiffuse += SpotAttenuationSpecular(i, WorldPos, WorldNormal, ViewDir);
		}
	}
	Out.OtherLights = TotalDiffuse;
		  
   return Out;
}


//-------------------------------------------------------------
// Pixel Shader (input channels):output channel
//-------------------------------------------------------------
float4 PS4(float4 Pos : POSITION0, float2 UV0 : TEXCOORD0, float3 ViewDir : TEXCOORD1, float3 Normal : TEXCOORD2, float3 Tangent : TEXCOORD3, float3 BiNormal : TEXCOORD4, float3 WorldPos : TEXCOORD5, float4 OtherLights : TEXCOORD6, float4 PosLightSpace : TEXCOORD7) : COLOR
{	
	float4 LightResult = 0;
	float3 NN = Normal;
	float4 Luminosity = float4(0,0,0,0);
	
	ViewDir = normalize(ViewDir);
		
	Normal = float3(0,0,1);
	if (b_UseNormal) 
	{
		Normal = 2.0f*tex2D(Normalmap, UV0.xy)-1.0f;
	}

	Normal = (Normal.x * Tangent) + (Normal.y * BiNormal) + (Normal.z * NN);	//Put normal into world space
	Normal = normalize(Normal);
		
	//Fresnel term
	half fresnel = saturate(dot(ViewDir, Normal));
	fresnel = pow(1-fresnel, fresnelPower);
	
	if (b_UseSpecMask)		// Use a texture for the specular mask
	{
		SpecMask = tex2D(Specularmap, UV0.xy);
	}
	SpecMask *= fresnel;	//Multiply the two terms to save multiplications

	if (b_UseLuminosity)		// Use a texture for the luminosity map
	{
		Luminosity = tex2D(Luminositymap, UV0.xy);
	}

	// Adding the vertexcolors lights
	float4 Color = OtherLights; 
	float3 Specular = OtherLights.aaa;
	
	for (int i = 0; i < iNumberOfLights && i<PIXELLIGHTS; i++)
	{
		if (light[i].iType==0)
		{  
			LightResult = DirectionalSpecular(i, Normal, ViewDir);
			Color.rgb += LightResult.rgb;
			Specular += LightResult.a*light[i].colour*SpecMask;
		}

		if (light[i].iType==1)
		{
			LightResult = OmniAttenuationSpecular(i, WorldPos, Normal, ViewDir);
			Color.rgb += LightResult.rgb;
			Specular += LightResult.a*light[i].colour*SpecMask;
		}

		if (light[i].iType==2 && light[i].bShadowActive)
		{ 	
			LightResult = SpotAttenuationSpecularShadow(i, WorldPos, Normal, ViewDir, PosLightSpace);
			Color.rgb += LightResult.rgb;
			Specular += LightResult.a*light[i].colour*SpecMask;	
		}
		if (light[i].iType==2 && light[i].bShadowActive == false)
		{ 	
			LightResult = SpotAttenuationSpecular(i, WorldPos, Normal, ViewDir);
			Color.rgb += LightResult.rgb;
			Specular += LightResult.a*light[i].colour*SpecMask;	
		}
	}

	if (b_UseAmbient)
	{
		half LerpFactor = (dot(Normal, Up) + 1) / 2;  
		Color.rgb += HemisphericColor*LerpFactor;
   	}

	Color.rgb += colourAmbient * fAmbientIntensity;

	if (Luminosity.x > 0.5f)
	{
		Color.rgb = float3(1,1,1);
	}

	if (b_UseDiffuse)
	{
		DiffuseColor = tex2D(Diffusemap, UV0.xy);
	}
 
	Color.rgb *= DiffuseColor.rgb;
	Color.rgb += Specular;

	if (b_UseReflection) 
	{
		float3 Reflections = texCUBE(EnvCubemap,-reflect(normalize(ViewDir),Normal))*SpecMask;
		Color.rgb += Reflections;
	}

	return float4(Color.rgb,1);
}


//-------------------------------------------------------------
// Name technique
//-------------------------------------------------------------
technique MaterialShader
{
	pass NoBones
	{
		// compiler directives
		VertexShader = compile vs_3_0 VS();
		ZEnable = true; 
		CullMode = CCW; 
		ZWriteEnable = true; 
		AlphaBlendEnable = false;
		PixelShader  = compile ps_3_0 PS4();
	}
}


